/**
 * 
 * File: FuzzyKNN.java
 * 
 * The Fuzzy K-NN algorithm. 
 * 
 * @author Written by Joaqun Derrac (University of Granada) 21/06/2011 
 * @version 1.0 
 * @since JDK1.5
 * 
 */

package keel.Algorithms.FingerprintClassification.Liu10;

import java.util.Arrays;
import java.util.StringTokenizer;

import org.core.Files;

import keel.Algorithms.FingerprintClassification.FingerprintClassificationAlgorithm;
import keel.Algorithms.FingerprintClassification.ReportTool;
import keel.Algorithms.FingerprintClassification.Timer;

public class Liu10 extends FingerprintClassificationAlgorithm {
	
	private static final int NUMSCALES = 4;
	
	private int nLeafs;
	private int nTrees;
	
	private int OVAOutput[];
	private Tree OVATrees [][];
	private double weights [][];
	private double sumLogLikehood [][];
	
	private boolean scales;
	private Tree OVATreesScales [][][];

	/** 
	 * Reads the parameters of the algorithm. 
	 * 
	 * @param script Configuration script
	 * 
	 */
	@Override
	protected void readParameters(String script) {
		
		String file;
		String line;
		String type;
		StringTokenizer fileLines, tokens;
		
	    file = Files.readFile (script);
	    fileLines = new StringTokenizer (file,"\n\r");
	    
	    //Discard in/out files definition
	    fileLines.nextToken();
	    fileLines.nextToken();
	    fileLines.nextToken();
	    
	    //Getting the number of leafs nodes per tree
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    nLeafs = Integer.parseInt(tokens.nextToken().substring(1));
	    
	    //Getting the number of trees per OVA
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    nTrees = Integer.parseInt(tokens.nextToken().substring(1));
	    
	    //Use of scales
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    type = tokens.nextToken().substring(1); 
	    
	    scales=false;
	    
	    if(type.equalsIgnoreCase("YES")){
	    	scales=true;
	    }
	    	

	} //end-method	

	/**
	 * Main builder. Initializes the methods' structures
	 * 
	 * @param script Configuration script
	 */
	public Liu10(String script){
		
		readDataFiles(script);
		
		//Naming the algorithm
		name="Liu10";
		
		OVAOutput = new int [trainOutput.length];
		OVATrees = new Tree [nClasses][nTrees];
		weights = new double [nClasses][trainOutput.length];
		sumLogLikehood = new double [nClasses][trainOutput.length];
		
		OVATreesScales = new Tree [NUMSCALES][nClasses][nTrees];
		
		Tree.setData(trainData);
		Tree.setLeafsLimit(nLeafs);
		Node.setDataSize(trainData.length);

	    //Initialization of Reporting tool
	    ReportTool.setOutputFile(outFile[2]);
	    
	} //end-method	
	
	/**
	 * Generates the model of the algorithm
	 */
	public void generateModel (){
		
		//Start of model time
		Timer.resetTime();	
		
		if(!scales){
			//build an OVA for each class
			for(int i=0;i<nClasses;i++){
				processTrainingOutput(i);
				buildOVA(i);
			}
		}
		else{
			for(int sc=0;sc<NUMSCALES;sc++){
				processTrainingData(sc);
				//build an OVA for each class
				for(int i=0;i<nClasses;i++){
					processTrainingOutput(i);
					buildOVAScales(i,sc);
				}
			}
		}
		
		//End of model time
		Timer.setModelTime();
	
		//Showing results
		System.out.println(name+" "+ relation + " Model " + Timer.getModelTime() + "s");
		
	}
	
	private void processTrainingData(int sc){
		
		double scaleData [][];
		
		scaleData= new double [trainData.length][inputAtt/NUMSCALES];
		
		switch(sc){
		
			case 0:
				for(int i=0;i<trainData.length;i++){
					System.arraycopy(trainData[i], 0, scaleData[i], 0, inputAtt/NUMSCALES);
				}
				
				break;
				
			case 1:
				for(int i=0;i<trainData.length;i++){
					System.arraycopy(trainData[i], inputAtt/NUMSCALES, scaleData[i], 0, inputAtt/NUMSCALES);
				}
				
				break;
				
			case 2:
				for(int i=0;i<trainData.length;i++){
					System.arraycopy(trainData[i], (2*inputAtt)/NUMSCALES, scaleData[i], 0, inputAtt/NUMSCALES);
				}
				
				break;
				
			case 3:
				for(int i=0;i<trainData.length;i++){
					System.arraycopy(trainData[i], (3*inputAtt)/NUMSCALES, scaleData[i], 0, inputAtt/NUMSCALES);
				}
				
				break;
				
			default:
				
		}
		
		Tree.setData(scaleData);

		
	}
	
	private void processTrainingOutput(int nClass){
		
		for(int i=0;i<OVAOutput.length;i++){
			if(trainOutput[i]==nClass){
				OVAOutput[i]=1;
			}
			else{
				OVAOutput[i]=-1;
			}
		}	
		
	}
	
	private void buildOVAScales(int nClass, int sc){
		
		Arrays.fill(weights[nClass], 1.0/(double)trainData.length);
		Arrays.fill(sumLogLikehood[nClass], 0.0);
		
		for(int t=0;t<nTrees; t++){
			
			OVATreesScales[sc][nClass][t]= new Tree(nClass, weights[nClass],OVAOutput);
			OVATreesScales[sc][nClass][t].generateRoot();
			
			while(OVATreesScales[sc][nClass][t].isFinished()==false){
				OVATreesScales[sc][nClass][t].expand();
			}
			
			System.out.println("New tree "+t+" "+nClass+" sc "+sc);
			OVATreesScales[sc][nClass][t].consolide();
			
			updateWeightsScales(nClass,t,sc);
			normalizeWeights(nClass);
		}
				
	}
	
	private void updateWeightsScales(int nClass,int tree,int sc){
		
		double factor;
		double acum;
		
		for(int i=0;i<trainData.length;i++){
			
			//add the LogLikehood ratios of the new tree
			sumLogLikehood[nClass][i]+=OVATreesScales[sc][nClass][tree].computeInstanceLogLikehood(trainData[i]);
			
			//compute the new weights
			acum=sumLogLikehood[nClass][i];
			factor=1.0+Math.exp(OVAOutput[i]*acum);
			weights[nClass][i]=weights[nClass][i]*(1.0/factor);
		}
		
	}
	
	private void buildOVA(int nClass){
		
		Arrays.fill(weights[nClass], 1.0/(double)trainData.length);
		Arrays.fill(sumLogLikehood[nClass], 0.0);
		
		for(int t=0;t<nTrees; t++){
			
			OVATrees[nClass][t]= new Tree(nClass, weights[nClass],OVAOutput);
			OVATrees[nClass][t].generateRoot();
			
			while(OVATrees[nClass][t].isFinished()==false){
				OVATrees[nClass][t].expand();
			}
			
			System.out.println("New tree "+t+" "+nClass);
			OVATrees[nClass][t].consolide();
			
			updateWeights(nClass,t);
			normalizeWeights(nClass);
		}
				
	}
	
	private void updateWeights(int nClass,int tree){
		
		double factor;
		double acum;
		
		for(int i=0;i<trainData.length;i++){
			
			//add the LogLikehood ratios of the new tree
			sumLogLikehood[nClass][i]+=OVATrees[nClass][tree].computeInstanceLogLikehood(trainData[i]);
			
			//compute the new weights
			acum=sumLogLikehood[nClass][i];
			factor=1.0+Math.exp(OVAOutput[i]*acum);
			weights[nClass][i]=weights[nClass][i]*(1.0/factor);
		}
		
	}
	
	private void normalizeWeights(int nClass){
		
		double sum=0.0;
		
		for(int i=0;i<weights[nClass].length;i++){
			sum+=weights[nClass][i];
		}
		
		for(int i=0;i<weights[nClass].length;i++){
			weights[nClass][i]/=sum;
		}
		
	}
	
	private int classifyInstance(double example[]){
		
		int output;
		double loglikehood[][];
		double sumClass[];
		
		loglikehood= new double [nClasses][nTrees];
		sumClass= new double [nClasses];
		
		//compute the loglikehood output over all trees
		for(int c=0;c<nClasses;c++){
			sumClass[c]=0.0;
			for(int t=0;t<nTrees;t++){
				loglikehood[c][t]=OVATrees[c][t].computeInstanceLogLikehood(example);

				//acum values over all trees
				sumClass[c]+=loglikehood[c][t];
			}
			
			//rescale loglikehood output using a logistic function [0,1]
			sumClass[c]=logistic(sumClass[c]);
	
		}
		
		//obtain the final class
		double max=sumClass[0];
		output=0;
		for(int c=1;c<nClasses;c++){
			if(sumClass[c]>max){
				max=sumClass[c];
				output=c;
			}
		}
		
		
		return output;
	}
	
	private int classifyInstanceScale(double example[]){
		
		int output;
		double loglikehood[][][];
		double sumClass[][];
		double pointsClass[];
		double exampleScale [];
		
		loglikehood= new double [NUMSCALES][nClasses][nTrees];
		sumClass= new double [NUMSCALES][nClasses];
		pointsClass= new double [nClasses];
		
		exampleScale= new double [inputAtt/NUMSCALES];
		
		//compute the loglikehood output over all trees and scales
		for(int sc=0;sc<NUMSCALES;sc++){
			
			System.arraycopy(example, (sc*inputAtt)/NUMSCALES, exampleScale, 0, inputAtt/NUMSCALES);
			
			for(int c=0;c<nClasses;c++){
				sumClass[sc][c]=0.0;
				for(int t=0;t<nTrees;t++){
					loglikehood[sc][c][t]=OVATreesScales[sc][c][t].computeInstanceLogLikehood(exampleScale);
	
					//acum values over all trees
					sumClass[sc][c]+=loglikehood[sc][c][t];
				}
				
				//rescale loglikehood output using a logistic function [0,1]
				sumClass[sc][c]=logistic(sumClass[sc][c]);
		
			}
		}
			
		//acum points per class
		for(int c=0;c<nClasses;c++){
			pointsClass[c]=0.0;
			for(int sc=0;sc<NUMSCALES;sc++){
				pointsClass[c]+=sumClass[sc][c];
			}
		}
		
		//obtain the final class
		double max=pointsClass[0];
		output=0;
		for(int c=1;c<nClasses;c++){
			if(pointsClass[c]>max){
				max=pointsClass[c];
				output=c;
			}
		}
		
		
		return output;
	}
	
	private double logistic(double a){
		
		double value;
		
		value=1.0/(1.0+Math.exp((-1.0*a)));
		
		return value;
	}
	
	/**
	 * Classifies the training set (leave-one-out)
	 */
	public void classifyTrain(){
	    
		//Start of training time
		Timer.resetTime();
		
		classifyTrainSet();
		
		//End of training time
		Timer.setTrainingTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Training " + Timer.getTrainingTime() + "s");
		
	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTest(){
		    
		//Start of training time
		Timer.resetTime();
		
		classifyTestSet();
		
		//End of test time
		Timer.setTestTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Test " + Timer.getTestTime() + "s");	
		
	} //end-method	
	
	/**
	 * Classifies the training set
	 */
	public void classifyTrainSet(){
				
		if(scales){
			for(int i=0;i<trainData.length;i++){
				
				trainPrediction[i]=classifyInstanceScale(trainData[i]);
				
			}
		}else{
			for(int i=0;i<trainData.length;i++){
							
				trainPrediction[i]=classifyInstance(trainData[i]);
				
			}
		}

	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTestSet(){

		if(scales){
			for(int i=0;i<testData.length;i++){
				
				testPrediction[i]=classifyInstanceScale(testData[i]);
				
			}
		}else{
			for(int i=0;i<testData.length;i++){
							
				testPrediction[i]=classifyInstance(testData[i]);
				
			}
		}

	} //end-method	
	
	/**
	 * Reports the results obtained
	 */
	public void printReport(){
		
		writeOutput(outFile[0], trainOutput, trainPrediction);
		writeOutput(outFile[1], testOutput, testPrediction);
		
		ReportTool.setResults(trainOutput,trainPrediction,testOutput,testPrediction,nClasses);
		
		ReportTool.printReport();

		
	} //end-method	
    
} //end-class 
