/**
 * 
 * File: FuzzyKNN.java
 * 
 * The Fuzzy K-NN algorithm. 
 * 
 * @author Written by Joaqun Derrac (University of Granada) 21/06/2011 
 * @version 1.0 
 * @since JDK1.5
 * 
 */

package keel.Algorithms.FingerprintClassification.Shah04;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.StringTokenizer;

import org.core.Files;
import org.core.Randomize;

import keel.Algorithms.FingerprintClassification.FingerprintClassificationAlgorithm;
import keel.Algorithms.FingerprintClassification.ReportTool;
import keel.Algorithms.FingerprintClassification.Timer;
import keel.Algorithms.FingerprintClassification.Util;


public class Shah04 extends FingerprintClassificationAlgorithm {
	
	private SMOWrapper SVM;
	private boolean selected [][][];
	
	private int firstClass,secondClass;
	private int votes1,votes2;
	
	private double C, gamma, epsilon;

	/** 
	 * Reads the parameters of the algorithm. 
	 * 
	 * @param script Configuration script
	 * 
	 */
	@Override
	protected void readParameters(String script) {
		
		String file;
		String line;
		String type;
		StringTokenizer fileLines, tokens;
		
	    file = Files.readFile (script);
	    fileLines = new StringTokenizer (file,"\n\r");
	    
	    //Discard in/out files definition
	    fileLines.nextToken();
	    fileLines.nextToken();
	    fileLines.nextToken();
	    
	    //Getting the C parameter
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    C = Double.parseDouble(tokens.nextToken().substring(1));
	    
	    //Getting the gamma parameter
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    gamma = Double.parseDouble(tokens.nextToken().substring(1));
	    
	    //Getting the epsilon parameter
	    line = fileLines.nextToken();
	    tokens = new StringTokenizer (line, "=");
	    tokens.nextToken();
	    epsilon  = Double.parseDouble(tokens.nextToken().substring(1));
	    
	   
  
	    
	} //end-method	

	/**
	 * Main builder. Initializes the methods' structures
	 * 
	 * @param script Configuration script
	 */
	public Shah04(String script){
		
		readDataFiles(script);
		
		//Naming the algorithm
		name="Shah04";

		SVM= new SMOWrapper(C,gamma,12345,epsilon);

		//Initialization of random generator    
	    Randomize.setSeed(seed);

	    //Initialization of Reporting tool
	    ReportTool.setOutputFile(outFile[2]);
	    
	    Timer.resetTime();
	    
	} //end-method	
	
	
	/**
	 * Generates the model of the algorithm
	 */
	public void generateModel (){
		
	
		try{
			SVM.generateTrainingFiles(trainFile);
			SVM.generateTestFiles(testFile);
		}catch(Exception e){}
		
		//obtain the support vectors
		SVM.classify();
		
		//fill arrays of instances
		selected = new boolean [nClasses][nClasses][trainData.length];
		
		int index;
		ArrayList<Integer> aux;
		for(int i=0;i<nClasses-1;i++){
			for(int j=i+1;j<nClasses;j++){
				Arrays.fill(selected[i][j], false);
				index=0;
				aux=SVM.getSupportVectors(i, j);
				for(int k=0;k<trainData.length;k++){
					if((trainOutput[k]==i)||(trainOutput[k]==j)){
						if(aux.contains(new Integer(index))){
							selected[i][j][k]=true;
						}
						index++;
					}
				}
			}
		}
		
		//End of model time
		Timer.setModelTime();
	
		//Showing results
		System.out.println(name+" "+ relation + " Model " + Timer.getModelTime() + "s");
		
	}
	
	
	/**
	 * Classifies the training set (leave-one-out)
	 */
	public void classifyTrain(){
	    
		//Start of training time
		Timer.resetTime();
		
		classifyTrainSet();
		
		//End of training time
		Timer.setTrainingTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Training " + Timer.getTrainingTime() + "s");
		
	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTest(){
		    
		//Start of training time
		Timer.resetTime();
		
		classifyTestSet();
		
		//End of test time
		Timer.setTestTime();
		
		//Showing results
		System.out.println(name+" "+ relation + " Test " + Timer.getTestTime() + "s");	
		
	} //end-method	
	
	/**
	 * Classifies the training set
	 */
	public void classifyTrainSet(){
				
		SVM.classify();
		
		for(int i=0;i<trainData.length;i++){
			
			trainPrediction[i]=classifyInstance(i,true);
			
		}

	} //end-method	
	
	/**
	 * Classifies the test set
	 */
	public void classifyTestSet(){

		for(int i=0;i<testData.length;i++){
			
			testPrediction[i]=classifyInstance(i,false);
			
		}

	} //end-method	
	
	private int classifyInstance(int index, boolean training){
		
		int output=-1;
		
		//kNN classification
		knnClassify(index,training);
		
		output=firstClass;
		
		if(votes2==0){
			return output;
		}
		
		//OVO NN classification
		if(training){
			if(firstClass<secondClass){
				output=SVMclassify(index,firstClass,secondClass,true);
	
			}
			else{
				output=SVMclassify(index,secondClass,firstClass,true);
			}
		}
		else{
			if(firstClass<secondClass){
				output=SVMclassify(index,firstClass,secondClass,false);
	
			}
			else{
				output=SVMclassify(index,secondClass,firstClass,false);
			}
		}
		return output;
	}
	
	private int SVMclassify(int index, int classA, int classB, boolean training){
		
		//NN is used, considering SVM support vectors
		double minDist;
		int output;
		double dist;
		
		minDist=Double.MAX_VALUE;
		output=-1;
	
		
		for (int i=0; i<trainData.length; i++) {
			
			if((!training)||(i!=index)){ //leave-one-out
				
				if(training){
					dist = Util.euclideanDistance(trainData[i],trainData[index]);
				}else{
					dist = Util.euclideanDistance(trainData[i],testData[index]);
				}
				
				if(minDist>dist){
					minDist=dist;
					output=trainOutput[i];
				}
			}
		}
		
		return output;
	}
	
	
	
	private void knnClassify(int index, boolean training){
		
		double minDist[];
		int nearestN[];
		double dist;
		boolean stop;
		int selectedClasses [];

		nearestN = new int[10];
		minDist = new double[10];
	
	    Arrays.fill(nearestN, -1);
	    Arrays.fill(minDist, Double.MAX_VALUE);
		
	    //KNN Method starts here
	    
		for (int i=0; i<trainData.length; i++) {
		
			if((!training)||(i!=index)){ //leave-one-out
				
				if(training){
					dist = Util.euclideanDistance(trainData[i],trainData[index]);
				}else{
					dist = Util.euclideanDistance(trainData[i],testData[index]);
				}
			    
	
				//see if it's nearer than our previous selected neighbors
				stop=false;
					
				for(int j=0;j<10 && !stop;j++){
					
					if (dist < minDist[j]) {
						    
						for (int l = 10 - 1; l >= j+1; l--) {
							minDist[l] = minDist[l - 1];
							nearestN[l] = nearestN[l - 1];
						}	
							
						minDist[j] = dist;
						nearestN[j] = i;
						stop=true;
					}
				}
			}
		}
		
		
		selectedClasses= new int[nClasses];
			
		Arrays.fill(selectedClasses, 0);	
				
		for (int i=0; i<10; i++) {
			selectedClasses[trainOutput[nearestN[i]]]+=1;
		}
				
		if(selectedClasses[0]>=selectedClasses[1]){
			firstClass=0;
			secondClass=1;
			votes1=selectedClasses[0];
			votes2=selectedClasses[1];
		}
		else{
			firstClass=1;
			secondClass=0;
			votes1=selectedClasses[1];
			votes2=selectedClasses[0];
			
		}
				
		for (int i=2; i<nClasses; i++) {
			if (votes1 < selectedClasses[i]) {
			
				votes2=votes1;
				votes1=selectedClasses[i];
				
				secondClass=firstClass;
				firstClass=i;
				
			}
			else{
				if (votes2 < selectedClasses[i]) {
					
					votes2=selectedClasses[i];
					
					secondClass=i;
				
				}
			}
		}

	}
	
	/**
	 * Reports the results obtained
	 */
	public void printReport(){
		
		writeOutput(outFile[0], trainOutput, trainPrediction);
		writeOutput(outFile[1], testOutput, testPrediction);
		
		ReportTool.setResults(trainOutput,trainPrediction,testOutput,testPrediction,nClasses);
		
		ReportTool.printReport();
		
	}
} //end-class 
